// ------------------------------------
// |            TileRogue             |
// | Copyright 2008 by Donnie Russell |
// |                                  |
// |           Version 1.21           |
// ------------------------------------



#include "rogue.h"



//640x480 or 800x600
#define CLIENTW  800
#define CLIENTH  600

#define TILEW  32
#define TILEH  32

//must not be larger than DCOLS
#define NUM_TILES_X  (CLIENTW/TILEW)

//must not be larger than DROWS
#if CLIENTH==600
  #define NUM_TILES_Y  12
#else
  #define NUM_TILES_Y  8
#endif

#define NUM_MSGS  ((CLIENTH-TILEH*NUM_TILES_Y-16)/8)



#define   DIR_BITMAP_FILE  "data/dir.dat"
#define  TILE_BITMAP_FILE  "data/tile.dat"
#define TITLE_BITMAP_FILE  "data/title.dat"
#define   WIN_BITMAP_FILE  "data/win.dat"
#define DEATH_BITMAP_FILE  "data/death.dat"

#define        SCORE_FILE  "scores.dat"



typedef struct {int w, h; unsigned short *p;} IMG;



#define COLOR16_R(c)  ((int)((((c)>>10)&31)<<3))
#define COLOR16_G(c)  ((int)((((c)>> 5)&31)<<3))
#define COLOR16_B(c)  ((int)(( (c)     &31)<<3))

#define COLOR16(r,g,b)  (((b)>>3)|(((g)>>3)<<5)|(((r)>>3)<<10))

//c1 shifted to c2 by a (0-256)
#define BLEND16(c1,c2,a) \
  (( (((c1)>>10)&31) +((( (((c2)>>10)&31) - (((c1)>>10)&31) )*(a))>>8) )<<10)| \
  (( (((c1)>> 5)&31) +((( (((c2)>> 5)&31) - (((c1)>> 5)&31) )*(a))>>8) )<< 5)| \
  (  ( (c1)     &31) +((( ( (c2)     &31) - ( (c1)     &31) )*(a))>>8)      )

#define    BACK_COLOR  COLOR16(  0,  0,  0)
#define   STAT1_COLOR  COLOR16(255,255,  0)
#define   STAT2_COLOR  COLOR16(255,255,255)
#define  PROMPT_COLOR  COLOR16(  0,255,255)
#define   INPUT_COLOR  COLOR16(  0,255,  0)
#define  NEWMSG_COLOR  COLOR16(128,255,128)
#define OLD1MSG_COLOR  COLOR16(192,255,192)
#define OLD2MSG_COLOR  COLOR16(120,180,120)
#define  CURSOR_COLOR  COLOR16(255,255,255)

#define BLACK     COLOR16(  0,  0,  0)
#define BLUE      COLOR16(  0,  0,128)
#define GREEN     COLOR16(  0,128,  0)
#define CYAN      COLOR16(  0,128,128)
#define RED       COLOR16(128,  0,  0)
#define MAGENTA   COLOR16(128,  0,128)
#define BROWN     COLOR16(128, 64,  0)
#define GREY      COLOR16(192,192,192)
#define DGREY     COLOR16(128,128,128)
#define LBLUE     COLOR16(  0,  0,255)
#define LGREEN    COLOR16(  0,255,  0)
#define LCYAN     COLOR16(  0,255,255)
#define LRED      COLOR16(255,  0,  0)
#define LMAGENTA  COLOR16(255,  0,255)
#define YELLOW    COLOR16(255,255,  0)
#define WHITE     COLOR16(255,255,255)

#define MASK_COLOR  COLOR16(255,255,255) //transparent color used in MaskedBlit



//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#define SetUpdateRegion(x1,x2,y1,y2)  do {                                \
  if ((x1)<UpdateX1) UpdateX1 = (x1); if ((x2)>UpdateX2) UpdateX2 = (x2); \
  if ((y1)<UpdateY1) UpdateY1 = (y1); if ((y2)>UpdateY2) UpdateY2 = (y2); \
  } while (0)



#define SetUpdateRegion_Empty()  do { \
  UpdateX1 = CLIENTW-1; UpdateX2 = 0; \
  UpdateY1 = CLIENTH-1; UpdateY2 = 0; \
  } while (0)



#define SetUpdateRegion_Full()  do {  \
  UpdateX1 = 0; UpdateX2 = CLIENTW-1; \
  UpdateY1 = 0; UpdateY2 = CLIENTH-1; \
  } while (0)
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



//////////////////////////////////////////////////////////////////////////////
//Geoff Kuenning's implementation of the Mersenne Twist pseudorandom number generator:
//http://www.cs.hmc.edu/~geoff/mtwist.html
void mt_seed32new(unsigned int seed);
unsigned int mt_lrand(void);



unsigned int RandInRange(unsigned int lower, unsigned int upper);
int get_rand(int x, int y);
int rand_percent(int percentage);
int coin_toss(void);

int LoadIMG_RLE(IMG *img, char *filename);
int LoadIMG_BMP(IMG *img, char *filename);
int SaveIMG_BMP(IMG *img, char *filename);
int AllocIMG(IMG *img, int w, int h);
void FreeIMG(IMG *img);
void ClearIMG(IMG *img);
void Put8x8Text(IMG *img, unsigned char *text, int x, int y, int color);
void Put8x16Text(IMG *img, unsigned char *text, int x, int y, int color);
void PutBox(int w, int h, int size_x, int size_y, int color, int *ret_x, int *ret_y);
void PutFilledRect(IMG *img, int x1, int y1, int x2, int y2, int color);
void Blit(IMG *si, int sx, int sy, IMG *di, int dx, int dy, int w, int h);
void MaskedBlit(IMG *si, int sx, int sy, IMG *di, int dx, int dy, int w, int h);
void BlendedBlit(IMG *si, int sx, int sy, IMG *di, int dx, int dy, int w, int h, int a);
void FilterRect(IMG *img, int x1, int y1, int x2, int y2, int type);
void FillNoiseIMG(IMG *img);
void RescaleHalf(IMG *si, IMG *di);

void SetDirArrow(int value);
void message(const char *p);

int ToggleSoundEnabled(void);
int get_input_line(char *prompt, char *insert, char *buf, char *if_cancelled, boolean add_blank, boolean do_echo);
int rgetchar(void);
int tick_pause(void);
void put_scores(const object *monster, short other);
void killed_by(const object *monster, short other);
void sound_bell(void);
void clean_up(char *message);
void print_help(void);
void draw_win(void);

void level_transition(void);



void PlayAudio(char *resource_name);
void UpdateClient(void);
void GetInput(int *code, int *ch);
int GetElapsed(int flag);
void InitInterface(void);
//////////////////////////////////////////////////////////////////////////////
