// ------------------------------------
// |            TileRogue             |
// | Copyright 2008 by Donnie Russell |
// |                                  |
// |           Version 1.21           |
// ------------------------------------



#include <stdlib.h> //malloc free atoi exit
#include <string.h> //memcpy strlen strcpy strcat
#include <time.h> //CLOCKS_PER_SEC clock
#include <math.h> //sqrt

#include "def.h"



//............................................................................
//0 before game starts and just before exiting program
//1 during game
//2 after game ends but before score is recorded
int WarningOnClose = 0;



int UpdateX1 = 0, UpdateX2 = CLIENTW-1, UpdateY1 = 0, UpdateY2 = CLIENTH-1;

int EnableFullScreen = 0;
int SoundEnabled = 1;

IMG ClientIMG = {0, 0, 0};

IMG   DarkIMG = {0, 0, 0};
IMG   TileIMG = {0, 0, 0};
IMG    DirIMG = {0, 0, 0};

int RenderWinFlag = 0;



extern unsigned char fnt_8x8_data[];
extern unsigned char fnt_8x16_data[];



extern int IMG_UpsideDownFlag;



extern int PrefixCount;

extern          short MovingObj_row       ;
extern          short MovingObj_col       ;
extern unsigned short MovingObj_which_kind;

extern fighter rogue;
extern object level_objects;
extern boolean wizard;
extern short max_level;
extern short cur_level;
extern short add_strength;
extern char login_name[MAX_OPT_LEN];
extern char fruit[MAX_OPT_LEN];
extern unsigned short dungeon[DROWS][DCOLS];
extern const char *const m_names[MONSTERS];
extern char hunger_str[HUNGER_STR_LEN];
extern char buffer[DCOLS*DROWS];
//............................................................................



//****************************************************************************
unsigned int RandInRange(unsigned int lower, unsigned int upper)
{
  return (((mt_lrand()&0x0000FFFFU)*(upper-lower+0x00000001U))>>16)+lower;
}



int get_rand(int x, int y)
{
  if (x>y) {int temp = y; y = x; x = temp;}

  return (int)RandInRange(x, y);
}



int rand_percent(int percentage)
{
  return (int)RandInRange(1, 100)<=percentage;
}



int coin_toss(void)
{
  return (int)RandInRange(0, 1);
}
//****************************************************************************



//****************************************************************************
//loads image files converted to rle format
int LoadIMG_RLE(IMG *img, char *filename)
{
  FILE *in;
  int w, h, i, n, offset;
  unsigned short *p;

  static unsigned short palette[256];

  in = fopen(filename, "rb"); if (in==0) return 0;

  if (fread(&w, sizeof(w), 1, in)==0) {fclose(in); return 0;}
  if (fread(&h, sizeof(h), 1, in)==0) {fclose(in); return 0;}

  for (i=0; i<256; i++) if (fread(&palette[i], sizeof(unsigned short), 1, in)==0) {fclose(in); return 0;}

  p = (unsigned short *)malloc(2*w*h); if (p==0) {fclose(in); return 0;}

  i = 0; n = 0; offset = 0;

  while (1)
  {
    while (n--)
    {
      if (IMG_UpsideDownFlag)     p[offset] = palette[i];
      else p[(h-1-(offset/w))*w+(offset%w)] = palette[i];

      if (++offset==w*h) goto done;
    }

    i = fgetc(in); if (i==EOF)           {free(p); fclose(in); return 0;}
    if (i==255)
    {
      n = fgetc(in); if (n==EOF || n==0) {free(p); fclose(in); return 0;}
      i = fgetc(in); if (i==EOF)         {free(p); fclose(in); return 0;}
    }
    else n = 1;
  }

done:
  fclose(in);

  img->w = w;
  img->h = h;
  img->p = p;

  return 1;
}



//loads 4 or 8 bpp palette bmp files
int LoadIMG_BMP(IMG *img, char *filename)
{
  unsigned short bfType;
  unsigned  long bfSize;
  unsigned short bfReserved1;
  unsigned short bfReserved2;
  unsigned  long bfOffBits;
  unsigned  long biSize;
            long biWidth;
            long biHeight;
  unsigned short biPlanes;
  unsigned short biBitCount;
  unsigned  long biCompression;
  unsigned  long biSizeImage;
            long biXPelsPerMeter;
            long biYPelsPerMeter;
  unsigned  long biClrUsed;
  unsigned  long biClrImportant;

  FILE *f;
  int r, g, b, x, y, i = 0, j = 0;
  unsigned short *p, *q;

  static unsigned short palette[256];

  f = fopen(filename, "rb"); if (f==0) return 0;

  fseek(f,  0, SEEK_SET); fread(&bfType         , 2, 1, f);
  fseek(f,  2, SEEK_SET); fread(&bfSize         , 4, 1, f);
  fseek(f,  6, SEEK_SET); fread(&bfReserved1    , 2, 1, f);
  fseek(f,  8, SEEK_SET); fread(&bfReserved2    , 2, 1, f);
  fseek(f, 10, SEEK_SET); fread(&bfOffBits      , 4, 1, f);
  fseek(f, 14, SEEK_SET); fread(&biSize         , 4, 1, f);
  fseek(f, 18, SEEK_SET); fread(&biWidth        , 4, 1, f);
  fseek(f, 22, SEEK_SET); fread(&biHeight       , 4, 1, f);
  fseek(f, 26, SEEK_SET); fread(&biPlanes       , 2, 1, f);
  fseek(f, 28, SEEK_SET); fread(&biBitCount     , 2, 1, f);
  fseek(f, 30, SEEK_SET); fread(&biCompression  , 4, 1, f);
  fseek(f, 34, SEEK_SET); fread(&biSizeImage    , 4, 1, f);
  fseek(f, 38, SEEK_SET); fread(&biXPelsPerMeter, 4, 1, f);
  fseek(f, 42, SEEK_SET); fread(&biYPelsPerMeter, 4, 1, f);
  fseek(f, 46, SEEK_SET); fread(&biClrUsed      , 4, 1, f);
  fseek(f, 50, SEEK_SET); fread(&biClrImportant , 4, 1, f);

  if (biBitCount!=4 && biBitCount!=8) {fclose(f); return 0;}

  p = (unsigned short *)malloc(2*biWidth*biHeight); if (p==0) {fclose(f); return 0;}

  fseek(f, 14+biSize, SEEK_SET);
  for (i=0; i<256; i++) {b = fgetc(f); g = fgetc(f); r = fgetc(f); fgetc(f); palette[i] = COLOR16(r,g,b);}

  fseek(f, bfOffBits, SEEK_SET);
  for (y=0; y<biHeight; y++)
  {
    if (IMG_UpsideDownFlag) q = p+biWidth*y; else q = p+biWidth*(biHeight-1-y);

    for (x=0; x<biWidth; x++)
    {
      if (biBitCount==4) {if ((x&1)==0) {j = fgetc(f); i = j>>4;} else i = j&15;}
      else if (biBitCount==8) i = fgetc(f);

      *q++ = palette[i];
    }

    while (x&3) {fgetc(f); x++;}
  }

  fclose(f);

  img->w = biWidth;
  img->h = biHeight;
  img->p = p;

  return 1;
}



//saves 16 bpp bmp files
int SaveIMG_BMP(IMG *img, char *filename)
{
  unsigned short bfType;
  unsigned  long bfSize;
  unsigned short bfReserved1;
  unsigned short bfReserved2;
  unsigned  long bfOffBits;
  unsigned  long biSize;
            long biWidth;
            long biHeight;
  unsigned short biPlanes;
  unsigned short biBitCount;
  unsigned  long biCompression;
  unsigned  long biSizeImage;
            long biXPelsPerMeter;
            long biYPelsPerMeter;
  unsigned  long biClrUsed;
  unsigned  long biClrImportant;

  FILE *f;
  int x, y;
  unsigned short *p;

  bfType = 'B'|('M'<<8);
  bfSize = 14+40+2*img->w*img->h;
  bfReserved1 = 0;
  bfReserved2 = 0;
  bfOffBits = 14+40;
  biSize = 40;
  biWidth = img->w;
  biHeight = img->h;
  biPlanes = 1;
  biBitCount = 16;
  biCompression = 0;
  biSizeImage = 0;
  biXPelsPerMeter = 0;
  biYPelsPerMeter = 0;
  biClrUsed = 0;
  biClrImportant = 0;

  f = fopen(filename, "wb"); if (f==0) return 0;

  fwrite(&bfType         , 2, 1, f);
  fwrite(&bfSize         , 4, 1, f);
  fwrite(&bfReserved1    , 2, 1, f);
  fwrite(&bfReserved2    , 2, 1, f);
  fwrite(&bfOffBits      , 4, 1, f);
  fwrite(&biSize         , 4, 1, f);
  fwrite(&biWidth        , 4, 1, f);
  fwrite(&biHeight       , 4, 1, f);
  fwrite(&biPlanes       , 2, 1, f);
  fwrite(&biBitCount     , 2, 1, f);
  fwrite(&biCompression  , 4, 1, f);
  fwrite(&biSizeImage    , 4, 1, f);
  fwrite(&biXPelsPerMeter, 4, 1, f);
  fwrite(&biYPelsPerMeter, 4, 1, f);
  fwrite(&biClrUsed      , 4, 1, f);
  fwrite(&biClrImportant , 4, 1, f);

  for (y=0; y<img->h; y++)
  {
    if (IMG_UpsideDownFlag) p = img->p+img->w*y; else p = img->p+img->w*(img->h-1-y);

    fwrite(p, 1, 2*img->w, f);

    x = img->w; while (x&3) {fputc(0, f); x++;}
  }

  fclose(f);

  return 1;
}



int AllocIMG(IMG *img, int w, int h)
{
  unsigned short *p;

  p = (unsigned short *)malloc(2*w*h); if (p==0) return 0;

  img->w = w;
  img->h = h;
  img->p = p;

  return 1;
}



void FreeIMG(IMG *img)
{
  if (img->p) {free(img->p); img->p = 0;}
}



void ClearIMG(IMG *img)
{
  memset(img->p, 0, (img->w*img->h)<<1);



  if (img==&ClientIMG) SetUpdateRegion_Full();
}

//----------------------------------------------------------------------------

#define PT \
  a = *f++; \
  if (a&128) d[0] = color; \
  if (a& 64) d[1] = color; \
  if (a& 32) d[2] = color; \
  if (a& 16) d[3] = color; \
  if (a&  8) d[4] = color; \
  if (a&  4) d[5] = color; \
  if (a&  2) d[6] = color; \
  if (a&  1) d[7] = color; \
  d += d1;



void Put8x8Text(IMG *img, unsigned char *text, int x, int y, int color)
{
  int start_x = x, d1, d2;
  unsigned short *d;

  d = img->p+img->w*(IMG_UpsideDownFlag?img->h-1-y:y)+x;
  d1 = (IMG_UpsideDownFlag?-1:1)*img->w;
  d2 = -(d1<<3)+8;

  while (*text && x+8<=img->w)
  {
    int ch = *text++;
    unsigned char *f = fnt_8x8_data+(ch<<3), a;

    PT PT PT PT PT PT PT PT

    x += 8; d += d2;
  }



  if (img==&ClientIMG) SetUpdateRegion(start_x, x-1, y, y+7);
}



void Put8x16Text(IMG *img, unsigned char *text, int x, int y, int color)
{
  int start_x = x, d1, d2;
  unsigned short *d;

  d = img->p+img->w*(IMG_UpsideDownFlag?img->h-1-y:y)+x;
  d1 = (IMG_UpsideDownFlag?-1:1)*img->w;
  d2 = -(d1<<4)+8;

  while (*text && x+8<=img->w)
  {
    int ch = *text++;
    unsigned char *f = fnt_8x16_data+(ch<<4), a;

    PT PT PT PT PT PT PT PT PT PT PT PT PT PT PT PT

    x += 8; d += d2;
  }



  if (img==&ClientIMG) SetUpdateRegion(start_x, x-1, y, y+15);
}



#undef PT



#define PB \
  if (h==8)  Put8x8Text(&ClientIMG, string, x+w*i, y, color); \
  else      Put8x16Text(&ClientIMG, string, x+w*i, y, color);



//w,h:            pixel size of one char
//size_x,size_y:  char size of box
//*ret_x,*ret_y:  pixel coordinates of upper-left corner of box
//                box is centered in tile area
void PutBox(int w, int h, int size_x, int size_y, int color, int *ret_x, int *ret_y)
{
  int x, y, i, j;
  unsigned char string[2] = "*";

  x = (          CLIENTW-w*size_x)>>1; if (x<0) x = 0;
  y = (TILEH*NUM_TILES_Y-h*size_y)>>1; if (y<0) y = 0;

  *ret_x = x;
  *ret_y = y;

  for (j=0; j<size_y; j++)
  {
    if (j==0 || j==size_y-1) for (i=0; i<size_x; i++)
    {
      string[0] = i==0 ? (j==0?218:192) : i==size_x-1 ? (j==0?191:217) : 196;

      PB
    }
    else for (i=0; i<size_x; i+=size_x-1)
    {
      string[0] = 179;

      PB
    }

    y += h;
  }
}



#undef PB

//----------------------------------------------------------------------------

void PutFilledRect(IMG *img, int x1, int y1, int x2, int y2, int color)
{
  int w, h, dd, i;
  unsigned short *d;

  if (x1<0) x1 = 0; if (x2>img->w-1) x2 = img->w-1;
  if (y1<0) y1 = 0; if (y2>img->h-1) y2 = img->h-1;

  w = x2-x1+1;
  h = y2-y1+1;

  if (w<=0 || h<=0) return;



  d = img->p+img->w*(IMG_UpsideDownFlag?img->h-1-y1:y1)+x1; dd = (IMG_UpsideDownFlag?-1:1)*img->w;

  while (h--)
  {
    for (i=0; i<w; i++) d[i] = color;
    d += dd;
  }



  if (img==&ClientIMG) SetUpdateRegion(x1, x2, y1, y2);
}

//----------------------------------------------------------------------------

#define BLIT_SETUP \
  unsigned short *s, *d; \
  int ss, dd; \
  if (dx<0) {w += dx; sx -= dx; dx = 0;} if (dx+w-di->w>0) w -= dx+w-di->w; if (w<=0 || sx<0 || sx+w-si->w>0) return; \
  if (dy<0) {h += dy; sy -= dy; dy = 0;} if (dy+h-di->h>0) h -= dy+h-di->h; if (h<=0 || sy<0 || sy+h-si->h>0) return; \
  s = si->p+si->w*(IMG_UpsideDownFlag?si->h-1-sy:sy)+sx; ss = (IMG_UpsideDownFlag?-1:1)*si->w; \
  d = di->p+di->w*(IMG_UpsideDownFlag?di->h-1-dy:dy)+dx; dd = (IMG_UpsideDownFlag?-1:1)*di->w; \
  if (di==&ClientIMG) SetUpdateRegion(dx, dx+w-1, dy, dy+h-1);



void Blit(IMG *si, int sx, int sy, IMG *di, int dx, int dy, int w, int h)
{
  BLIT_SETUP

  while (h--)
  {
    memcpy(d, s, w<<1);

    s += ss;
    d += dd;
  }
}



void MaskedBlit(IMG *si, int sx, int sy, IMG *di, int dx, int dy, int w, int h)
{
  BLIT_SETUP

  while (h--)
  {
    int i;

    for (i=0; i<w; i++) if (s[i]!=MASK_COLOR) d[i] = s[i];

    s += ss;
    d += dd;
  }
}



//a: 0-256
void BlendedBlit(IMG *si, int sx, int sy, IMG *di, int dx, int dy, int w, int h, int a)
{
  BLIT_SETUP

  while (h--)
  {
    int i;

    for (i=0; i<w; i++) if (s[i]!=COLOR16(0,0,0)) d[i] = BLEND16(d[i],s[i],a);

    s += ss;
    d += dd;
  }
}

//----------------------------------------------------------------------------

//apply rgb color filter specified by type to rectangle
void FilterRect(IMG *img, int x1, int y1, int x2, int y2, int type)
{
  int w, h, r, g, b, dd, i;
  unsigned short *d;

  if (x1<0) x1 = 0; if (x2>img->w-1) x2 = img->w-1;
  if (y1<0) y1 = 0; if (y2>img->h-1) y2 = img->h-1;

  w = x2-x1+1;
  h = y2-y1+1;

  if (w<=0 || h<=0) return;



  r = 0; g = 0; b = 0; switch (type)
  {
    case 1: r = 0; g = 2; b = 2; break;
    case 2: r = 2; g = 0; b = 2; break;
    case 3: r = 2; g = 2; b = 0; break;
    case 4: r = 2; g = 2; b = 2; break;
  }
  r += 10; g += 5;



  d = img->p+img->w*(IMG_UpsideDownFlag?img->h-1-y1:y1)+x1; dd = (IMG_UpsideDownFlag?-1:1)*img->w;



  while (h--)
  {
    for (i=0; i<w; i++) d[i] = ((d[i]&31)>>b) | (((d[i]&(31<<5))>>g)<<5) | (((d[i]&(31<<10))>>r)<<10);

    d += dd;
  }



  if (img==&ClientIMG) SetUpdateRegion(x1, x2, y1, y2);
}

//----------------------------------------------------------------------------

//seed random number generator before calling this function
void FillNoiseIMG(IMG *img)
{
  unsigned short *p;
  int n, i;

  p = img->p; n = img->w*img->h; while (n--)
  {
    i = RandInRange(0, 99);
         if (i<70) i = RandInRange(256*0/16, 256*2/16);
    else if (i<97) i = RandInRange(256*2/16, 256*3/16);
    else           i = RandInRange(256*3/16, 256*4/16);

    *p++ = COLOR16(i>>1,i>>1,i);
  }



  if (img==&ClientIMG) SetUpdateRegion_Full();
}

//----------------------------------------------------------------------------

//uses box filter, best suited for 1/2 rescaling
//for this function, di must be exactly 1/2 dimensions of si
void RescaleHalf(IMG *si, IMG *di)
{
  int j;

  for (j=0; j<di->h; j++)
  {
    unsigned short *d = di->p+di->w*j;
    int i;

    for (i=0; i<di->w; i++)
    {
      int x = i<<1, y = j<<1;
      int c0 = *(si->p+si->w*(y+0)+x+0);
      int c1 = *(si->p+si->w*(y+1)+x+0);
      int c2 = *(si->p+si->w*(y+0)+x+1);
      int c3 = *(si->p+si->w*(y+1)+x+1);

      *d++ = COLOR16((COLOR16_R(c0)+COLOR16_R(c1)+COLOR16_R(c2)+COLOR16_R(c3))>>2,
                     (COLOR16_G(c0)+COLOR16_G(c1)+COLOR16_G(c2)+COLOR16_G(c3))>>2,
                     (COLOR16_B(c0)+COLOR16_B(c1)+COLOR16_B(c2)+COLOR16_B(c3))>>2);
    }
  }



  if (di==&ClientIMG) SetUpdateRegion_Full();
}
//****************************************************************************



//****************************************************************************
int DirArrow = -1;



//-1 no arrow   0-7 one arrow   8 all arrows
void SetDirArrow(int value)
{
  DirArrow = value;
}

//----------------------------------------------------------------------------

int BonesX = 0;
int BonesY = 0;
int BonesType = 0;



//calculate bones location and type from door coordinates
void CalcBones(void)
{
  int i, x, y;

  static short old_cur_level = -1;

  if (cur_level==old_cur_level) return;
  old_cur_level = cur_level;

  for (i=1; i<=3; i++)
  {
    BonesX = BonesY = 0;
    for (y=0; y<DROWS; y++) for (x=0; x<DCOLS; x++) if (dungeon[y][x]&DOOR)
    {
      BonesX += x*i;
      BonesY += y*i;
    }
    BonesX %= DCOLS;
    BonesY %= DROWS;

    if ((dungeon[BonesY][BonesX]&FLOOR) || (dungeon[BonesY][BonesX]&TUNNEL)) break;
  }

  BonesType = (BonesX+BonesY)%3; //3 bones tiles
}

//----------------------------------------------------------------------------

//occurrences of (?<<5) assume TILEW and TILEH are 32



#define TILE_WALL      ( 0*TILEW)
#define TILE_FLOOR     ( 1*TILEW)
#define TILE_TUNNEL    ( 2*TILEW)
#define TILE_STAIRS    ( 3*TILEW)
#define TILE_TRAP      ( 4*TILEW)
#define TILE_DOOR      ( 5*TILEW)
#define TILE_ROGUE     ( 6*TILEW)
#define TILE_MONSTERA  ( 7*TILEW)
#define TILE_WEAPON0   (33*TILEW)
#define TILE_ARMOR0    (41*TILEW)
#define TILE_GOLD      (48*TILEW)
#define TILE_FOOD      (49*TILEW)
#define TILE_SCROLL    (50*TILEW)
#define TILE_POTION    (51*TILEW)
#define TILE_RING      (52*TILEW)
#define TILE_STAFF     (53*TILEW)
#define TILE_AMULET    (54*TILEW)
#define TILE_BONES0    (55*TILEW)
#define TILE_FIRE      (58*TILEW)
#define TILE_ICE       (59*TILEW)



#define PLOT_DARK         {Blit(&DarkIMG, (x<<5)&255, (y<<5)&255, &ClientIMG, sx, sy, TILEW, TILEH);}
#define PLOT_TILE(tile)   {Blit(&TileIMG, (tile), 0, &ClientIMG, sx, sy, TILEW, TILEH);}
#define PLOT_MTILE(tile)  {MaskedBlit(&TileIMG, (tile), 0, &ClientIMG, sx, sy, TILEW, TILEH);}
#define PLOT_DIRARROW     {if (DirArrow>=0 && DirArrow<=8) MaskedBlit(&DirIMG, DirArrow<<5, 0, &ClientIMG, sx, sy, TILEW, TILEH);}



void RenderLargeMap(void)
{
  int vx, vy, tx, ty, x, y, sx, sy, flags, ch;

  CalcBones();

  vx = rogue.col-NUM_TILES_X/2; if (vx>DCOLS-NUM_TILES_X) vx = DCOLS-NUM_TILES_X; if (vx<0) vx = 0;
  vy = rogue.row-NUM_TILES_Y/2; if (vy>DROWS-NUM_TILES_Y) vy = DROWS-NUM_TILES_Y; if (vy<0) vy = 0;

  for (ty=0; ty<NUM_TILES_Y; ty++)
  for (tx=0; tx<NUM_TILES_X; tx++)
  {
    x = vx+tx; sx = tx<<5;
    y = vy+ty; sy = ty<<5;

    ch = buffer[y*DCOLS+x]; if (ch==' ') {PLOT_DARK continue;}

    flags = dungeon[y][x];
    if (flags&FLOOR)
    {
                                                  PLOT_TILE(TILE_FLOOR)
      if (flags&STAIRS)                           PLOT_MTILE(TILE_STAIRS)
      else if ((flags&TRAP) && (flags&HIDDEN)==0) PLOT_MTILE(TILE_TRAP)
      else if (x==BonesX && y==BonesY)            PLOT_MTILE(TILE_BONES0+(BonesType<<5))
    }
    else if (flags&TUNNEL)
    {
                                                  PLOT_TILE(TILE_TUNNEL)
      if (flags&STAIRS)                           PLOT_MTILE(TILE_STAIRS)
      else if ((flags&TRAP) && (flags&HIDDEN)==0) PLOT_MTILE(TILE_TRAP)
      else if (x==BonesX && y==BonesY)            PLOT_MTILE(TILE_BONES0+(BonesType<<5))
    }
    else if (flags&HORWALL)                       PLOT_TILE(TILE_WALL)
    else if (flags&VERTWALL)                      PLOT_TILE(TILE_WALL)
    else if (flags&DOOR)
    {
                                                  PLOT_TILE(TILE_WALL)
      if ((flags&HIDDEN)==0)                      PLOT_MTILE(TILE_DOOR)
    }

    if (ch>='A' && ch<='Z') PLOT_MTILE(TILE_MONSTERA+((ch-'A')<<5))
    else switch (ch)
    {
      case '@': PLOT_MTILE(TILE_ROGUE) PLOT_DIRARROW break;

      case ')':
      {
        int tile = TILE_WEAPON0;
        object *obj = level_objects.next_object;

        if (MovingObj_row==y && MovingObj_col==x && MovingObj_which_kind<WEAPONS) tile += (MovingObj_which_kind)<<5;
        else while (obj)
        {
          if (obj->row==y && obj->col==x && obj->which_kind<WEAPONS) {tile += (obj->which_kind)<<5; break;}
          obj = obj->next_object;
        }

        PLOT_MTILE(tile)
        break;
      }

      case ']':
      {
        int tile = TILE_ARMOR0;
        object *obj = level_objects.next_object;

        if (MovingObj_row==y && MovingObj_col==x && MovingObj_which_kind<ARMORS) tile += (MovingObj_which_kind)<<5;
        else while (obj)
        {
          if (obj->row==y && obj->col==x && obj->which_kind<ARMORS) {tile += (obj->which_kind)<<5; break;}
          obj = obj->next_object;
        }

        PLOT_MTILE(tile)
        break;
      }

      case '*': PLOT_MTILE(TILE_GOLD  )               break;
      case ':': PLOT_MTILE(TILE_FOOD  )               break;
      case '?': PLOT_MTILE(TILE_SCROLL)               break;
      case '!': PLOT_MTILE(TILE_POTION)               break;
      case '=': PLOT_MTILE(TILE_RING  )               break;
      case '/': PLOT_MTILE(TILE_STAFF )               break;
      case ',': PLOT_MTILE(TILE_AMULET)               break;

      case '0': PLOT_MTILE(TILE_FIRE  )               break;
      case '1': PLOT_MTILE(TILE_ICE   )               break;
    }
  }
}



#define PLOT(color) \
  { \
    if (IMG_UpsideDownFlag) *d = *(d+1) = *(d-CLIENTW) = *(d-CLIENTW+1) = *(d-CLIENTW*2) = *(d-CLIENTW*2+1) = color; \
    else                    *d = *(d+1) = *(d+CLIENTW) = *(d+CLIENTW+1) = *(d+CLIENTW*2) = *(d+CLIENTW*2+1) = color; \
  }



void RenderSmallMap(void)
{
  unsigned short *d;
  int x, y, dd, flags, ch;
  char *p;

  x = CLIENTW-2*DCOLS;
  y = TILEH*NUM_TILES_Y+16;



  SetUpdateRegion(x, x+2*DCOLS-1, y, y+3*DROWS-1);



  d = ClientIMG.p+CLIENTW*(IMG_UpsideDownFlag?CLIENTH-1-y:y)+x;
  dd = (IMG_UpsideDownFlag?-1:1)*CLIENTW*3-2*DCOLS;



  p = buffer;
  for (y=0; y<DROWS; y++,d+=dd)
  for (x=0; x<DCOLS; x++,d+=2)
  {
    ch = *p++; if (ch==' ') {PLOT(DGREY) continue;}

    flags = dungeon[y][x];
    if (flags&FLOOR)
    {
                                                  PLOT(GREY)
      if (flags&STAIRS)                           PLOT(MAGENTA)
      else if ((flags&TRAP) && (flags&HIDDEN)==0) PLOT(BLUE)
    }
    else if (flags&TUNNEL)
    {
                                                  PLOT(GREY)
      if (flags&STAIRS)                           PLOT(MAGENTA)
      else if ((flags&TRAP) && (flags&HIDDEN)==0) PLOT(BLUE)
    }
    else if (flags&HORWALL)                       PLOT(WHITE)
    else if (flags&VERTWALL)                      PLOT(WHITE)
    else if (flags&DOOR)
    {
                                                  PLOT(WHITE)
      if ((flags&HIDDEN)==0)                      PLOT(GREY)
    }

    if (ch>='A' && ch<='Z') PLOT(LRED)
    else switch (ch)
    {
      case '@': PLOT(BLACK ) break;
      case ')': PLOT(GREEN ) break;
      case ']': PLOT(GREEN ) break;
      case '*': PLOT(YELLOW) break;
      case ':': PLOT(GREEN ) break;
      case '?': PLOT(GREEN ) break;
      case '!': PLOT(GREEN ) break;
      case '=': PLOT(GREEN ) break;
      case '/': PLOT(GREEN ) break;
      case ',': PLOT(GREEN ) break;

      case '0': PLOT(LRED  ) break;
      case '1': PLOT(WHITE ) break;
    }
  }
}

//----------------------------------------------------------------------------

#define STATUS_WORD_SIZE  128
#define NUM_STATUS_WORDS  8



void RenderStatusLine(int forget_next_old)
{
  int x, y, i;

  static char new_words[STATUS_WORD_SIZE*NUM_STATUS_WORDS];
  static char old_words[STATUS_WORD_SIZE*NUM_STATUS_WORDS];

  static int first_time = 1;
  static int forget_old = 0;

  if (first_time==0 && forget_old==0) memcpy(old_words, new_words, STATUS_WORD_SIZE*NUM_STATUS_WORDS);



  sprintf(new_words+STATUS_WORD_SIZE*0, "Level:%d", cur_level);
  sprintf(new_words+STATUS_WORD_SIZE*1, "Gold:%ld", rogue.gold);
  sprintf(new_words+STATUS_WORD_SIZE*2, "Hp:%d(%d)", rogue.hp_current, rogue.hp_max);
  sprintf(new_words+STATUS_WORD_SIZE*3, "Str:%d(%d)", rogue.str_current+add_strength, rogue.str_max);
  sprintf(new_words+STATUS_WORD_SIZE*4, "Arm:%d", get_armor_class(rogue.armor));
  sprintf(new_words+STATUS_WORD_SIZE*5, "Exp:%d/%ld", rogue.exp, rogue.exp_points);
  sprintf(new_words+STATUS_WORD_SIZE*6, "%s", hunger_str);
  if (PrefixCount) sprintf(new_words+STATUS_WORD_SIZE*7, "Count:%d", PrefixCount); else sprintf(new_words+STATUS_WORD_SIZE*7, "");



  if (first_time) memcpy(old_words, new_words, STATUS_WORD_SIZE*NUM_STATUS_WORDS);

  x = 0; y = TILEH*NUM_TILES_Y;
  for (i=0; i<NUM_STATUS_WORDS; i++)
  {
    char *p = new_words+STATUS_WORD_SIZE*i;
    int color = STAT2_COLOR;

    if (*p)
    {
      if (strcmp(p, old_words+STATUS_WORD_SIZE*i)==0) color = STAT1_COLOR;
      if (i) x += 8*2;
      Put8x16Text(&ClientIMG, (unsigned char *)p, x, y, color);
      x += 8*strlen(p);
    }
  }

  first_time = 0;
  if (forget_next_old) forget_old = 1; else forget_old = 0;
}

//----------------------------------------------------------------------------

#define MSG_SIZE  128



char Msgs[MSG_SIZE*NUM_MSGS];

int NumUnviewedMsgs = 0;



void message(const char *p)
{
  int i;

  for (i=0; i<NUM_MSGS-1; i++) memcpy(Msgs+MSG_SIZE*i, Msgs+MSG_SIZE*(i+1), MSG_SIZE);
  strncpy(Msgs+MSG_SIZE*(NUM_MSGS-1), p, 79);

  NumUnviewedMsgs++;
}



void RenderMsgs(int forget_next_old)
{
  int i, color;

  static int num = 0;

  for (i=0; i<NUM_MSGS; i++)
  {
         if (i>=NUM_MSGS-NumUnviewedMsgs    ) color =  NEWMSG_COLOR;
    else if (i>=NUM_MSGS-NumUnviewedMsgs-num) color = OLD1MSG_COLOR;
    else                                      color = OLD2MSG_COLOR;

    Put8x8Text(&ClientIMG, (unsigned char *)(Msgs+MSG_SIZE*i), 0, TILEH*NUM_TILES_Y+16+(i<<3), color);
  }

  if (forget_next_old==0 && NumUnviewedMsgs)
  {
    num = NumUnviewedMsgs;
    NumUnviewedMsgs = 0;
  }
}

//----------------------------------------------------------------------------

void RenderWin(void)
{
  IMG img = {0, 0, 0};
  int y;

  FilterRect(&ClientIMG, 0, 0, CLIENTW-1, TILEH*NUM_TILES_Y-1, 4);

  if (LoadIMG_RLE(&img, WIN_BITMAP_FILE))
  {
    y = (TILEH*NUM_TILES_Y-img.h)>>1;
    MaskedBlit(&img, 0, 0, &ClientIMG, (CLIENTW-img.w)>>1, y, img.w, img.h);
    y += img.h;
    y += (TILEH*NUM_TILES_Y-y-16*1)>>1;

    FreeIMG(&img);
  }
  else y = (TILEH*NUM_TILES_Y-16*1)>>1;

  Put8x16Text(&ClientIMG, (unsigned char *)"Press Space to view scores", (CLIENTW-8*26)>>1, y, WHITE);
}

//----------------------------------------------------------------------------

void Render(int forget_next_old)
{
  ClearIMG(&ClientIMG);

  RenderLargeMap();
  RenderSmallMap();
  RenderStatusLine(forget_next_old);
  RenderMsgs(forget_next_old);

  if (RenderWinFlag) RenderWin();
}
//****************************************************************************



//****************************************************************************
int ToggleSoundEnabled(void)
{
  SoundEnabled = 1-SoundEnabled;

  return SoundEnabled;
}



#define CURSOR_BLINK_DELAY  500 //milliseconds



#define GIL_CLEAR   PutFilledRect(&ClientIMG, 0, y, CLIENTW-1, y+16-1, BACK_COLOR);
#define GIL_CURSOR  PutFilledRect(&ClientIMG, (n+1+i)*8, y, (n+1+i)*8+7, y+16-1, cursor_on?CURSOR_COLOR:BACK_COLOR);
#define GIL_PROMPT  Put8x16Text(&ClientIMG, (unsigned char *)prompt, 0, y, PROMPT_COLOR);
#define GIL_INPUT   Put8x16Text(&ClientIMG, (unsigned char *)buf, (n+1)*8, y, INPUT_COLOR);



int get_input_line(char *prompt, char *insert, char *buf, char *if_cancelled, boolean add_blank, boolean do_echo)
{
  int y, n, i, first_time, cursor_on, code, ch;

  y = TILEH*NUM_TILES_Y;
  n = strlen(prompt);
  i = 0; if (insert[0]) {strcpy(buf, insert); i = strlen(insert);}
  buf[i] = 0;

  while (1)
  {
    GIL_CLEAR  GIL_PROMPT  if (do_echo) GIL_INPUT

    first_time = 1; cursor_on = 0;
    while (1)
    {
      if (first_time || GetElapsed(1)>=CURSOR_BLINK_DELAY)
      {
        first_time = 0; cursor_on ^= 1;
        GIL_CURSOR  UpdateClient();
        GetElapsed(0);
      }

      GetInput(&code, &ch); if (code || ch) break;
    }

    if (ch=='\r' || ch==CANCEL) break;
    if (ch>=' ' && ch<='~' && i<MAX_TITLE_LENGTH-2) {buf[i++] = ch; buf[i] = 0;}
    if (ch=='\b' && i>0) {i--; buf[i] = 0;}
  }

  GIL_CLEAR  UpdateClient();

  if (add_blank) {buf[i++] = ' '; buf[i] = 0;}
  if (ch==CANCEL || i==0 || (i==1 && add_blank!=0))
  {
    if (if_cancelled) message(if_cancelled);
    return 0;
  }
  return i;
}



int rgetchar(void)
{
  int code, ch;

  Render(0);
  UpdateClient();

  while (1)
  {
    GetInput(&code, &ch);

    switch (code)
    {
      case 71: ch = 'y'; break;
      case 72: ch = 'k'; break;
      case 73: ch = 'u'; break;
      case 75: ch = 'h'; break;
      case 77: ch = 'l'; break;
      case 79: ch = 'b'; break;
      case 80: ch = 'j'; break;
      case 81: ch = 'n'; break;

      case 2071: ch = 'Y'; break;
      case 2072: ch = 'K'; break;
      case 2073: ch = 'U'; break;
      case 2075: ch = 'H'; break;
      case 2077: ch = 'L'; break;
      case 2079: ch = 'B'; break;
      case 2080: ch = 'J'; break;
      case 2081: ch = 'N'; break;
    }

    if (ch) break;
  }

  return ch;
}



//return: 1 esc pressed
int tick_pause(void)
{
  int i, code, ch;

  Render(1);
  UpdateClient();

  i = 2; while (i--) {GetInput(&code, &ch); if (ch==27) return 1;}

  return 0;
}



void FilterByOther(short other)
{
  int filter_type = 4;

  switch (other)
  {
    case        QUIT: filter_type = 4; break;

    case HYPOTHERMIA: filter_type = 3; break;
    case  STARVATION: filter_type = 4; break;
    case POISON_DART: filter_type = 2; break;
    case       KFIRE: filter_type = 1; break;

    case           0: filter_type = 1; break;

    case         WIN: filter_type = 4; break;
  }

  FilterRect(&ClientIMG, 0, 0, CLIENTW-1, TILEH*NUM_TILES_Y-1, filter_type);
}



#define NUM_SCORES   10
#define SCORE_WIDTH  128



void put_scores(const object *monster, short other)
{
  FILE *f;
  char scores[NUM_SCORES][SCORE_WIDTH];
  int rank, i;

  //fill entries with spaces
  for (rank=0; rank<NUM_SCORES; rank++)
  {
    for (i=0; i<SCORE_WIDTH-1; i++) scores[rank][i] = ' ';
    scores[rank][i] = 0;
  }

  //read score file
  f = fopen(SCORE_FILE, "rb");
  if (f)
  {
    for (rank=0; rank<NUM_SCORES; rank++) if (fread(scores[rank], 1, SCORE_WIDTH, f)<SCORE_WIDTH) break;
    fclose(f);
  }

  //find position of new entry
  for (rank=0; rank<NUM_SCORES; rank++) if (rogue.gold>=atoi(scores[rank]+6)) break;

  //if entry in top position
  if (rank<NUM_SCORES)
  {
    //open space for new entry
    for (i=NUM_SCORES-1; i>rank; i--) strcpy(scores[i], scores[i-1]);

    //create new entry
    sprintf(scores[rank], "      %-6ld  %s: ", rogue.gold, login_name);
    switch (other)
    {
      case        QUIT: strcat(scores[rank], "quit"); break;

      case HYPOTHERMIA: strcat(scores[rank], "died of hypothermia"); break;
      case  STARVATION: strcat(scores[rank], "died of starvation"); break;
      case POISON_DART: strcat(scores[rank], "killed by a dart"); break;
      case       KFIRE: strcat(scores[rank], "killed by fire"); break;

      case 0:
        strcat(scores[rank], "killed by ");
        if (is_vowel(m_names[monster->m_char-'A'][0])) strcat(scores[rank], "an "); else strcat(scores[rank], "a ");
        strcat(scores[rank], m_names[monster->m_char-'A']);
      break;

      case WIN:         strcat(scores[rank], "a total winner"); break;
    }
    sprintf(scores[rank]+strlen(scores[rank]), " on level %d ", max_level);
    if (other!=WIN && has_amulet()) strcat(scores[rank], "with amulet");

    //pad entry with spaces
    for (i=strlen(scores[rank]); i<SCORE_WIDTH-1; i++) scores[rank][i] = ' ';
    scores[rank][i] = 0;
  }

  //put rank numbers
  for (i=0; i<NUM_SCORES; i++)
  {
    if ((i+1)<10) {scores[i][0] = '0'+(i+1); scores[i][1] = ' ';}
    else {scores[i][0] = '0'+(i+1)/10; scores[i][1] = '0'+(i+1)%10;}
  }

  //write score file
  f = fopen(SCORE_FILE, "wb");
  if (f)
  {
    for (i=0; i<NUM_SCORES; i++) if (fwrite(scores[i], 1, SCORE_WIDTH, f)<SCORE_WIDTH) break;
    fclose(f);
  }



  WarningOnClose = 0;



  Render(1);

  FilterByOther(other);



  {
    int maxlen, x, y;
    char *heading = "Rank  Score   Name";

    maxlen = strlen(heading);

    //strip spaces from end of each scores string
    for (i=0; i<NUM_SCORES; i++)
    {
      int len = strlen(scores[i]);

      while (--len>=0) if (scores[i][len]!=' ') break;
      scores[i][++len] = 0;

      if (len>maxlen) maxlen = len;
    }



    PutBox(8, 16, 1+maxlen+1, 1+1+NUM_SCORES+1, LBLUE, &x, &y);

    Put8x16Text(&ClientIMG, (unsigned char *)heading, x+8*1, y+16*1, WHITE);
    for (i=0; i<NUM_SCORES; i++) Put8x16Text(&ClientIMG, (unsigned char *)scores[i], x+8*1, y+16*(2+i), GREY);

    Put8x16Text(&ClientIMG, (unsigned char *)"Press Space to exit", (CLIENTW-8*19)>>1, y+16*(1+1+NUM_SCORES+1), WHITE);

    while (1)
    {
      int color, code, ch;

      color = clock()%CLOCKS_PER_SEC;
      if (color>=CLOCKS_PER_SEC/2) color = CLOCKS_PER_SEC-1-color;
      color = 256-80+color*80/(CLOCKS_PER_SEC/2);
      color = COLOR16(0,color,0);

      Put8x16Text(&ClientIMG, (unsigned char *)scores[rank], x+8*1, y+16*(2+rank), color);

      UpdateClient();

      GetInput(&code, &ch); if (ch==' ') break;
    }
  }



  exit(0);
}



#define SKULLFADE_MULT  10 //larger: slower



void killed_by(const object *monster, short other)
{
  WarningOnClose = 2;



  if (other!=QUIT)
  {
    IMG A_IMG = {0, 0, 0};
    IMG B_IMG = {0, 0, 0};
    IMG C_IMG = {0, 0, 0};
    IMG img;
    char buf[128];
    int x, y, i, code, ch;



    PlayAudio("RogueDieSound");

    rogue.gold = (rogue.gold*9)/10;



    strcpy(buf, login_name);

    switch (other)
    {
      case HYPOTHERMIA: strcat(buf, ", died of hypothermia"); break;
      case  STARVATION: strcat(buf, ", died of starvation"); break;
      case POISON_DART: strcat(buf, ", killed by a dart"); break;
      case       KFIRE: strcat(buf, ", killed by fire"); break;

      case 0:
        strcat(buf, ", killed by ");
        if (is_vowel(m_names[monster->m_char-'A'][0])) strcat(buf, "an "); else strcat(buf, "a ");
        strcat(buf, m_names[monster->m_char-'A']);
      break;

      default: strcat(buf, ", died mysteriously"); break;
    }



    if (LoadIMG_RLE(&A_IMG, DEATH_BITMAP_FILE)==0) goto done;
    if (AllocIMG(&B_IMG, A_IMG.w>>1, A_IMG.h>>1)==0) goto done;
    img = TILEH*NUM_TILES_Y-A_IMG.h-16*2>=0 ? A_IMG : B_IMG;
    if (AllocIMG(&C_IMG, img.w, img.h)==0) goto done;

    x = (          CLIENTW-img.w)>>1; if (x<0) x = 0;
    y = (TILEH*NUM_TILES_Y-img.h)>>1; if (y<0) y = 0;

    RescaleHalf(&A_IMG, &B_IMG);

    Render(1);
    FilterByOther(other);
    Blit(&ClientIMG, x, y, &C_IMG, 0, 0, img.w, img.h);

    Put8x16Text(&ClientIMG, (unsigned char *)buf, (CLIENTW-8*strlen(buf))>>1, y+img.h, GREY);
    Put8x16Text(&ClientIMG, (unsigned char *)"Press Space to view scores", (CLIENTW-8*26)>>1, y+img.h+16, WHITE);

    i = 0*SKULLFADE_MULT; while (i<256*SKULLFADE_MULT)
    {
      GetElapsed(0);

      Blit(&C_IMG, 0, 0, &ClientIMG, x, y, img.w, img.h);
      BlendedBlit(&img, 0, 0, &ClientIMG, x, y, img.w, img.h, i/SKULLFADE_MULT);

      UpdateClient();
      GetInput(&code, &ch); //throw away keystrokes
      i += GetElapsed(1);
    }

    while (1) {GetInput(&code, &ch); if (ch==' ') break;}

done:
    FreeIMG(&A_IMG);
    FreeIMG(&B_IMG);
    FreeIMG(&C_IMG);
  }



  put_scores(monster, other);
}



void sound_bell(void)
{
  PlayAudio("BellSound");
}



//don't Render() in this function because game data may not be initialized
void clean_up(char *message)
{
  WarningOnClose = 0;



  {
    int x, y;

    FilterRect(&ClientIMG, 0, 0, CLIENTW-1, TILEH*NUM_TILES_Y-1, 4);

    PutBox(8, 16, 1+strlen(message)+1, 1+1+1, LBLUE, &x, &y);
    Put8x16Text(&ClientIMG, (unsigned char *)message, x+8*1, y+16*1, GREY);
    Put8x16Text(&ClientIMG, (unsigned char *)"Press Space to exit", (CLIENTW-8*19)>>1, y+16*3, WHITE);

    UpdateClient();
  }



  while (1)
  {
    int code, ch;

    GetInput(&code, &ch); if (ch==' ') break;
  }



  exit(0);
}



char *HelpString[] =
{
  "y  k  u   SHIFT (dir)  run in direction",
  " \\ | /    CTRL  (dir)  run until adjacent",
  "h -+- l   ESC          cancel command",
  " / | \\    *            list all objects at prompt",
  "b  j  n",
  "",
  "c (obj)   call object by name              t (dir) (obj)  throw object",
  "d (obj)   drop object                      T              take off armor",
  "e (obj)   eat food                         v              print version number",
  "f (dir)   fight until death or near death  w (obj)        wield weapon",
  "F (dir)   fight until one opponent dies    W (obj)        wear armor",
  "i         list all objects                 z (dir) (obj)  zap wand",
  "m (dir)   move onto without picking up     ,              pick up object",
  "P (obj)   put on ring                      <num> .        rest <num> turns",
  "q (obj)   quaff potion                     >              go down staircase",
  "Q         quit                             <              go up staircase",
  "r (obj)   read scroll                      )              print current weapon",
  "R         remove ring                      ]              print current armor",
  "<num> s   search area <num> times          =              print current rings",
  "S         save game                        ^ (dir)        identify trap type",
  "",
  "CTRL s    toggle sound effects on/off",
  "_",
  "_CTRL c    create inventory object          CTRL r         reveal level map",
  "_CTRL i    list objects on level            CTRL t         reveal traps on",
  "_CTRL m    reveal monsters on level map                      level map",
  "_CTRL o    reveal objects on level map      CTRL w         turn off wizard mode",

  0
};



#define GET_HELP_STRING  HelpString[i]+(HelpString[i][0]=='_'?1:0)



void print_help(void)
{
  int maxlen, num, i, w, h, x, y, step, pos, code, ch;

  maxlen = 0; num = 0; i = 0; while (HelpString[i])
  {
    if (wizard || HelpString[i][0]!='_')
    {
      char *p = GET_HELP_STRING;
      int len = strlen(p);

      if (len>maxlen) maxlen = len;
      num++;
    }
    i++;
  }

  if (maxlen==0 || num==0) return;



  FilterRect(&ClientIMG, 0, 0, CLIENTW-1, TILEH*NUM_TILES_Y-1, 4);
  PutFilledRect(&ClientIMG, 0, TILEH*NUM_TILES_Y, CLIENTW-1, TILEH*NUM_TILES_Y+15, BLACK);
  Put8x16Text(&ClientIMG, (unsigned char *)"Press any key to continue", (CLIENTW-8*25)/2, TILEH*NUM_TILES_Y, WHITE);

  w = 8; if (TILEH*NUM_TILES_Y-16*num<0) h = 8; else h = 16;

  PutBox(w, h, 1+maxlen+1, 1+num+1, LBLUE, &x, &y);



  for (step=0; step<num; step++)
  {
    pos = 0; i = 0; while (HelpString[i])
    {
      if (wizard || HelpString[i][0]!='_')
      {
        char *p = GET_HELP_STRING;
        int color;

        color = pos-(num>>1); if (color<0) color = -color;
        color = (color?num/color:num)+step; if (color>num) color = num;
        color = 192*color/num;
        color = COLOR16(color,color,color);

        if (h==16) Put8x16Text(&ClientIMG, (unsigned char *)p, x+w*1, y+h*(1+pos), color);
        else        Put8x8Text(&ClientIMG, (unsigned char *)p, x+w*1, y+h*(1+pos), color);

        pos++;
      }
      i++;
    }

    UpdateClient();

    GetInput(&code, &ch); if (code || ch) return;
  }



  while (1)
  {
    GetInput(&code, &ch); if (code || ch) return;
  }
}



char *CopyrightString[] =
{
  "Copyright (c) 1988, 1993 The Regents of the University of California.  All rights reserved.",
  "This code is derived from software contributed to Berkeley by Timothy C. Stoehr.",
  "",
  "Redistribution and use in source and binary forms, with or without modification, are permitted",
  "provided that the following conditions are met:",
  "1. Redistributions of source code must retain the above copyright notice, this list of conditions",
  "   and the following disclaimer.",
  "2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions",
  "   and the following disclaimer in the documentation and/or other materials provided with the",
  "   distribution.",
  "3. All advertising materials mentioning features or use of this software must display the following",
  "   acknowledgement: This product includes software developed by the University of California,",
  "   Berkeley and its contributors.",
  "4. Neither the name of the University nor the names of its contributors may be used to endorse or",
  "   promote products derived from this software without specific prior written permission.",
  "",
  "THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY EXPRESS OR IMPLIED",
  "WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR",
  "A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR",
  "ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT",
  "LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS",
  "INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,",
  "OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN",
  "IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.",

  0
};



void draw_copyright(int clear_flag)
{
  int maxlen = 0, num = 0, i;

  while (CopyrightString[num])
  {
    int len = strlen(CopyrightString[num]);

    if (len>maxlen) maxlen = len;
    num++;
  }

  if (maxlen==0 || maxlen>CLIENTW/8 || num==0 || num>NUM_MSGS) return;



  for (i=0; i<num; i++) Put8x8Text(&ClientIMG, (unsigned char *)CopyrightString[i], 0, CLIENTH-8*num+8*i, clear_flag?BLACK:DGREY);
}



void draw_title(void)
{
  int x, y;
  IMG img = {0, 0, 0};

  for (y=0; y<NUM_TILES_Y; y++)
  for (x=0; x<NUM_TILES_X; x++)
  {
    if (x==0 || x==NUM_TILES_X-1 || y==0 || y==NUM_TILES_Y-1) Blit(&TileIMG, TILE_WALL, 0, &ClientIMG, x<<5, y<<5, TILEW, TILEH);
    else Blit(&DarkIMG, (x<<5)&255, (y<<5)&255, &ClientIMG, x<<5, y<<5, TILEW, TILEH);
  }

  if (LoadIMG_RLE(&img, TITLE_BITMAP_FILE))
  {
    MaskedBlit(&img, 0, 0, &ClientIMG, (CLIENTW-img.w)>>1, (TILEH*NUM_TILES_Y-img.h)>>1, img.w, img.h);

    FreeIMG(&img);
  }
}



void draw_win(void)
{
  Render(1);
  RenderWinFlag = 1;
  level_transition();
  Render(1); UpdateClient(); //in case transition failed
  RenderWinFlag = 0;

  while (1)
  {
    int code, ch;

    GetInput(&code, &ch); if (ch==' ') break;
  }
}
//****************************************************************************



//****************************************************************************
#define ELLIPSE_CXL  (0)
#define ELLIPSE_CXH  (CLIENTW-1)
#define ELLIPSE_CYL  (0)
#define ELLIPSE_CYH  (TILEH*NUM_TILES_Y-1)

#define ELLIPSE_BITS  9
#define ELLIPSE_RAD  (1<<ELLIPSE_BITS)



short EllipseData[2*ELLIPSE_RAD+1];



void InitCopyEllipse(void)
{
  int i;

  static int init = 0;

  if (!init)
  {
    for (i=-ELLIPSE_RAD; i<=ELLIPSE_RAD; i++) EllipseData[i+ELLIPSE_RAD] = (short)sqrt((double)(ELLIPSE_RAD*ELLIPSE_RAD-i*i));
    init = 1;
  }
}



//assumes x and y (center) are inside clip region
//assumes rx and ry (radii) are >= 0
//src and dest must be same size
void CopyEllipse(IMG *si, IMG *di, int x, int y, int rx, int ry)
{
  int pos = 0, step = ((ELLIPSE_RAD*2+1)<<16)/((ry<<1)+1), yl = y-ry, yh = y+ry;
  int dd, i, j;
  unsigned short *s, *d;

  if (yl<ELLIPSE_CYL) {pos += (ELLIPSE_CYL-yl)*step; yl = ELLIPSE_CYL;}
  if (yh>ELLIPSE_CYH) yh = ELLIPSE_CYH;



  i = di->w*(IMG_UpsideDownFlag?di->h-1-yl:yl); dd = (IMG_UpsideDownFlag?-1:1)*di->w;

  s = si->p+i;
  d = di->p+i;



  j = yh-yl+1; while (j-->0)
  {
    int dx = (rx*EllipseData[pos>>16])>>ELLIPSE_BITS, xl = x-dx, xh = x+dx;

    if (xl<ELLIPSE_CXL) xl = ELLIPSE_CXL;
    if (xh>ELLIPSE_CXH) xh = ELLIPSE_CXH;

    i = (xh-xl+1)<<1; if (i>0) memcpy(d+xl, s+xl, i);

    pos += step; s += dd; d += dd;
  }



  {
    int xl = x-rx, xh = x+rx;

    if (xl<ELLIPSE_CXL) xl = ELLIPSE_CXL;
    if (xh>ELLIPSE_CXH) xh = ELLIPSE_CXH;

    SetUpdateRegion(xl, xh, yl, yh);
  }
}

//----------------------------------------------------------------------------

#define TRANSITION_MULT  3 //larger: slower

#define TRANSITION_MAX_SIZE  (CLIENTW/2+48)



void level_transition(void)
{
  IMG A_IMG = {0, 0, 0};
  IMG B_IMG = {0, 0, 0};
  int i, code, ch;

  if (AllocIMG(&A_IMG, CLIENTW, CLIENTH)==0) goto done;
  if (AllocIMG(&B_IMG, CLIENTW, CLIENTH)==0) goto done;

  memcpy(A_IMG.p, ClientIMG.p, 2*CLIENTW*CLIENTH);
  Render(1);
  memcpy(B_IMG.p, ClientIMG.p, 2*CLIENTW*CLIENTH);
  memcpy(ClientIMG.p, A_IMG.p, 2*CLIENTW*CLIENTH);



  InitCopyEllipse();

  i = 0*TRANSITION_MULT; while (i<TRANSITION_MAX_SIZE*TRANSITION_MULT)
  {
    GetElapsed(0);

    CopyEllipse(&B_IMG, &ClientIMG, CLIENTW/2, (TILEH*NUM_TILES_Y)/2, i/TRANSITION_MULT, i/TRANSITION_MULT);
    UpdateClient();

    GetInput(&code, &ch); //throw away keystrokes

    i += GetElapsed(1);
  }



done:
  FreeIMG(&A_IMG);
  FreeIMG(&B_IMG);
}
//****************************************************************************



//############################################################################
void FreeIMGs(void)
{
  FreeIMG(&DarkIMG);
  FreeIMG(&TileIMG);
  FreeIMG(& DirIMG);
}



int wrapped_main(int argc, char **argv)
{
  int i;

  for (i=1; i<argc; i++)
  {
    if (!strcmp(argv[i], "-full_screen")) EnableFullScreen = 1;
  }



  InitInterface();



  mt_seed32new((unsigned int)time(0));

  atexit(FreeIMGs);

  if (AllocIMG(&DarkIMG, 256, 256)==0) exit(1);
  FillNoiseIMG(&DarkIMG);

  if (LoadIMG_RLE(&TileIMG, TILE_BITMAP_FILE)==0) exit(1);
  if (LoadIMG_RLE(& DirIMG,  DIR_BITMAP_FILE)==0) exit(1);



  draw_title();
  draw_copyright(0); //draw
  if (!get_input_line("Rogue's name:",   "", login_name, 0, 0, 1)) strcpy(login_name, "nobody");
  if (!get_input_line("Favorite fruit:", "", fruit,      0, 1, 1)) strcpy(fruit, "slime-mold ");
  draw_copyright(1); //clear



  WarningOnClose = 1;



  game(); //never returns
  return 0; //never reached
}
//############################################################################
