// ------------------------------------
// |            TileRogue             |
// | Copyright 2008 by Donnie Russell |
// |                                  |
// |           Version 1.21           |
// ------------------------------------



#include <SDL/SDL.h>

#include <SDL_mixer.h>

#include "def.h"



//............................................................................
int IMG_UpsideDownFlag = 0;



int ActiveFlag = 1;
int ScrollLockFlag = 0;



SDL_Surface *ClientSurface = 0;
SDL_Surface *MemSurface = 0;



#define NUM_RESOURCES  19

struct {char *r; char *f;} ResourceToFilename[NUM_RESOURCES] =
{
  {"EatSound",         "sounds/eat.wav"      },
  {"RogueHitSound",    "sounds/hit.wav"      },
  {"RogueMissSound",   "sounds/miss.wav"     },
  {"MonsterDieSound",  "sounds/kill.wav"     },
  {"MonsterHitSound",  "sounds/hit2.wav"     },
  {"MonsterMissSound", "sounds/miss2.wav"    },
  {"ExperienceSound",  "sounds/level.wav"    },
  {"StairsSound",      "sounds/stairs.wav"   },
  {"GoldSound",        "sounds/gold.wav"     },
  {"ZapSound",         "sounds/zap.wav"      },
  {"TeleportSound",    "sounds/teleport.wav" },
  {"TrapSound",        "sounds/trap.wav"     },
  {"RogueDieSound",    "sounds/harp.wav"     },
  {"WinSound",         "sounds/win.wav"      },
  {"BellSound",        "sounds/bell.wav"     },
  {"ArmorSound",       "sounds/armor.wav"    },
  {"ConfusionSound",   "sounds/confusion.wav"},
  {"FireSound",        "sounds/fire.wav"     },
  {"IceSound",         "sounds/ice.wav"      }
};



extern int UpdateX1, UpdateX2, UpdateY1, UpdateY2;

extern int EnableFullScreen;
extern int SoundEnabled;

extern IMG ClientIMG;
//............................................................................



//****************************************************************************
void TranslateSym(SDLKey sym, SDLMod mod, Uint16 unicode, int *code, int *ch)
{
  *code = 0; *ch = 0;

  if (unicode>0 && unicode<128) {*ch = unicode; return;}



  switch (sym)
  {
    case SDLK_HOME    : case SDLK_KP7      : *code = 71; break;
    case SDLK_UP      : case SDLK_KP8      : *code = 72; break;
    case SDLK_PAGEUP  : case SDLK_KP9      : *code = 73; break;
    case SDLK_LEFT    : case SDLK_KP4      : *code = 75; break;
    case SDLK_RIGHT   : case SDLK_KP6      : *code = 77; break;
    case SDLK_END     : case SDLK_KP1      : *code = 79; break;
    case SDLK_DOWN    : case SDLK_KP2      : *code = 80; break;
    case SDLK_PAGEDOWN: case SDLK_KP3      : *code = 81; break;
    case SDLK_INSERT  : case SDLK_KP0      : *code = 82; break;
    case SDLK_DELETE  : case SDLK_KP_PERIOD: *code = 83; break;

    case SDLK_F1 : *code = 59; break;
    case SDLK_F2 : *code = 60; break;
    case SDLK_F3 : *code = 61; break;
    case SDLK_F4 : *code = 62; break;
    case SDLK_F5 : *code = 63; break;
    case SDLK_F6 : *code = 64; break;
    case SDLK_F7 : *code = 65; break;
    case SDLK_F8 : *code = 66; break;
    case SDLK_F9 : *code = 67; break;
    case SDLK_F10: *code = 68; break;
  }

  if (mod&KMOD_ALT   && *code>=59 && *code<=68) *code += 1000;

  if (mod&KMOD_SHIFT && *code>=71 && *code<=83) *code += 2000;



  if (sym==SDLK_SCROLLOCK) ScrollLockFlag = 1-ScrollLockFlag;
}
//****************************************************************************



//****************************************************************************
void PlayAudio(char *resource_name)
{
  int i;
  char *filename;
  Mix_Chunk *chunk;

  if (SoundEnabled==0) return;



  for (i=0; i<NUM_RESOURCES; i++) if (strcmp(resource_name, ResourceToFilename[i].r)==0) break;
  if (i==NUM_RESOURCES) return;
  filename = ResourceToFilename[i].f;

  chunk = Mix_LoadWAV(filename);

  Mix_PlayChannel(-1, chunk, 0);
}
//****************************************************************************



//****************************************************************************
void UpdateClient(void)
{
  int w, h;

  if (UpdateX1<0) UpdateX1 = 0; if (UpdateX2>CLIENTW-1) UpdateX2 = CLIENTW-1;
  if (UpdateY1<0) UpdateY1 = 0; if (UpdateY2>CLIENTH-1) UpdateY2 = CLIENTH-1;

  w = UpdateX2-UpdateX1+1;
  h = UpdateY2-UpdateY1+1;

  if (w>0 && h>0)
  {
    SDL_Rect r;

    r.x = (Sint16)UpdateX1; r.w = (Uint16)w;
    r.y = (Sint16)UpdateY1; r.h = (Uint16)h;

    SDL_BlitSurface(MemSurface, &r, ClientSurface, &r);
    SDL_UpdateRect(ClientSurface, r.x, r.y, r.w, r.h);
  }

  SetUpdateRegion_Empty();
}



void SetShellMode(int mode)
{
  SDL_WM_SetCaption(mode?"Command Prompt":"TileRogue", mode?"Command Prompt":"TileRogue");
}



void GetInput(int *code, int *ch)
{
  SDL_Event event;



  SDL_Delay(2);



  *code = 0;
  *ch = 0;

  while (SDL_PollEvent(&event)) switch (event.type)
  {
    case SDL_QUIT:        exit(0);
    case SDL_KEYDOWN:     TranslateSym(event.key.keysym.sym, event.key.keysym.mod, event.key.keysym.unicode, code, ch); break;
    case SDL_KEYUP:       break;
    case SDL_ACTIVEEVENT: if (event.active.state&SDL_APPACTIVE) {if (event.active.gain) ActiveFlag = 1; else ActiveFlag = 0;} break;
  }



  if (*code==1062) exit(0); //exit program if [alt f4] pressed
}



void GetLockStates(int *scroll, int *num, int *caps)
{
  unsigned char *keystate;
  int numkeys;

  keystate = SDL_GetKeyState(&numkeys);

  *scroll = ScrollLockFlag;
  *num    = keystate[SDLK_NUMLOCK ]!=0;
  *caps   = keystate[SDLK_CAPSLOCK]!=0;
}



//SDL_GetTicks(): number of milliseconds since SDL library initialization, up to 49.7 days
int GetElapsed(int flag)
{
  unsigned long e;

  static unsigned long start;

  if (!flag) {start = SDL_GetTicks(); return 0;}

  e = SDL_GetTicks()-start; if (e<1) e = 1; if (e>1000*10) e = 1000*10;

  return (int)e;
}



void ShutdownInterface(void)
{
  if (MemSurface) {SDL_FreeSurface(MemSurface); MemSurface = 0;}

  FreeIMG(&ClientIMG);

  Mix_CloseAudio();
  SDL_Quit();
}



void InitInterface(void)
{
  if (SDL_Init(SDL_INIT_VIDEO|SDL_INIT_TIMER|SDL_INIT_AUDIO)<0) {fprintf(stderr, "SDL_Init() failed: %s\n", SDL_GetError()); exit(1);}
  if (Mix_OpenAudio(MIX_DEFAULT_FREQUENCY, MIX_DEFAULT_FORMAT, 2, 1024)<0) {SDL_Quit(); fprintf(stderr, "Mix_OpenAudio() failed"); exit(1);}

  atexit(ShutdownInterface);

  if (EnableFullScreen) ClientSurface = SDL_SetVideoMode(CLIENTW, CLIENTH, 16, SDL_HWSURFACE|SDL_FULLSCREEN);
  else                  ClientSurface = SDL_SetVideoMode(CLIENTW, CLIENTH,  0, SDL_HWSURFACE               );

  if (ClientSurface==0) {fprintf(stderr, "SDL_SetVideoMode() failed: %s\n", SDL_GetError()); exit(1);}

  if (AllocIMG(&ClientIMG, CLIENTW, CLIENTH)==0) exit(1);

  MemSurface = SDL_CreateRGBSurfaceFrom(ClientIMG.p, ClientIMG.w, ClientIMG.h, 16, 2*ClientIMG.w, 31<<10, 31<<5, 31<<0, 0);
  if (MemSurface==0) {fprintf(stderr, "SDL_CreateRGBSurfaceFrom() failed: %s\n", SDL_GetError()); exit(1);}

  SDL_ShowCursor(EnableFullScreen?SDL_DISABLE:SDL_ENABLE);

  SDL_WM_SetCaption("TileRogue", "TileRogue");

  SDL_EnableUNICODE(1);
  SDL_EnableKeyRepeat(250, 25);



  ClearIMG(&ClientIMG);
  UpdateClient();
}
//**************************************************************************



//###########################################################################
int wrapped_main(int argc, char **argv);



int main(int argc, char **argv)
{
  return wrapped_main(argc, argv);
}
//###########################################################################
