#include <stdio.h>
#include <stdlib.h>
#include <string.h>



//palette index 255 should be avoided, if possible,
//because it costs 3 bytes for run lengths of 1



#define RLE_INIT   last_i = -1; n = 0;

#define RLE \
  {if (last_i!=i || n==255) \
  { \
    if (n>1 || last_i==255) {fputc(255, out); fputc(n, out); fputc(last_i, out); n = 0;} \
    else if (n==1) {fputc(last_i, out); n = 0;} \
  } \
  last_i = i; n++;}

#define RLE_END  i = -1; RLE



#define COLOR16(r,g,b)  (((b)>>3)|(((g)>>3)<<5)|(((r)>>3)<<10))



//converts 4 or 8 bpp palette bmp files to rle (run-length encoded) format
int BMP_To_RLE(char *filename_in, char *filename_out)
{
  unsigned short bfType;
  unsigned long  bfSize;
  unsigned short bfReserved1;
  unsigned short bfReserved2;
  unsigned long  bfOffBits;
  unsigned long  biSize;
           long  biWidth;
           long  biHeight;
  unsigned short biPlanes;
  unsigned short biBitCount;
  unsigned long  biCompression;
  unsigned long  biSizeImage;
           long  biXPelsPerMeter;
           long  biYPelsPerMeter;
  unsigned long  biClrUsed;
  unsigned long  biClrImportant;

  FILE *in, *out;
  int x, y, i, last_i, n, j = 0;

  in  = fopen(filename_in , "rb"); if ( in==0) return 0;
  out = fopen(filename_out, "wb"); if (out==0) {fclose(in); return 0;}

  fseek(in,  0, SEEK_SET); fread(&bfType         , 2, 1, in);
  fseek(in,  2, SEEK_SET); fread(&bfSize         , 4, 1, in);
  fseek(in,  6, SEEK_SET); fread(&bfReserved1    , 2, 1, in);
  fseek(in,  8, SEEK_SET); fread(&bfReserved2    , 2, 1, in);
  fseek(in, 10, SEEK_SET); fread(&bfOffBits      , 4, 1, in);
  fseek(in, 14, SEEK_SET); fread(&biSize         , 4, 1, in);
  fseek(in, 18, SEEK_SET); fread(&biWidth        , 4, 1, in);
  fseek(in, 22, SEEK_SET); fread(&biHeight       , 4, 1, in);
  fseek(in, 26, SEEK_SET); fread(&biPlanes       , 2, 1, in);
  fseek(in, 28, SEEK_SET); fread(&biBitCount     , 2, 1, in);
  fseek(in, 30, SEEK_SET); fread(&biCompression  , 4, 1, in);
  fseek(in, 34, SEEK_SET); fread(&biSizeImage    , 4, 1, in);
  fseek(in, 38, SEEK_SET); fread(&biXPelsPerMeter, 4, 1, in);
  fseek(in, 42, SEEK_SET); fread(&biYPelsPerMeter, 4, 1, in);
  fseek(in, 46, SEEK_SET); fread(&biClrUsed      , 4, 1, in);
  fseek(in, 50, SEEK_SET); fread(&biClrImportant , 4, 1, in);

  if (biBitCount!=4 && biBitCount!=8) {fclose(out); fclose(in); return 0;}

  fwrite(&biWidth , sizeof(biWidth ), 1, out);
  fwrite(&biHeight, sizeof(biHeight), 1, out);

  fseek(in, 14+biSize, SEEK_SET);
  for (i=0; i<256; i++)
  {
    int r, g, b;
    unsigned short color;

    b = fgetc(in); g = fgetc(in); r = fgetc(in); fgetc(in);
    color = (unsigned short)COLOR16(r,g,b);
    fwrite(&color, sizeof(color), 1, out);
  }
  fseek(in, bfOffBits, SEEK_SET);



  RLE_INIT
  for (y=0; y<biHeight; y++)
  {
    for (x=0; x<biWidth; x++)
    {
      if (biBitCount==4) {if ((x&1)==0) {j = fgetc(in); i = j>>4;} else i = j&15;} else if (biBitCount==8) i = fgetc(in);
      RLE
    }
    while (x&3) {fgetc(in); x++;}
  }
  RLE_END



  fclose(out);
  fclose(in);

  return 1;
}



int main(int argc, char *argv[])
{
  int i, flag = 0, num_errors = 0;

  for (i=1; i<argc; i++)
  {
    if (!strncmp(argv[i], "-e", 2) && argc>i+2)
    {
      if (BMP_To_RLE(argv[i+1], argv[i+2])==0) {printf("Failed to convert %s to %s.\n", argv[i+1], argv[i+2]); num_errors++;}
      else printf("%s converted to %s.\n", argv[i+1], argv[i+2]);

      flag = 1; i += 2;
    }
  }

  if (!flag) {printf("%s   -e input1.bmp output1.dat   -e input2.bmp output2.dat   ...\n", argv[0]); return 0;}

  if (num_errors) {printf("  There were %s errors.\n", num_errors); return 1;}

  printf("  All files converted successfully.\n"); return 0;
}
