/*
 * Copyright (c) 1988, 1993
 *  The Regents of the University of California.  All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * Timothy C. Stoehr.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *  This product includes software developed by the University of
 *  California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 *
 * This source herein may be modified and/or distributed by anybody who
 * so desires, with the following restrictions:
 *    1.)  No portion of this notice shall be removed.
 *    2.)  Credit shall not be taken for the creation of this source.
 *    3.)  This code is not to be traded, sold, or used for personal
 *         gain or profit.
 *
 */

//The following source code is derived from the NetBSD version of Rogue
//(last modified 2002/10/01)



#include <stdio.h> //FILE fopen fseek SEEK_SET fread fclose fgetc fwrite fputc
#include <time.h> //time



#define boolean  char

#define swap(x,y)  {t = x; x = y; y = t;}

#define NOTHING   ((unsigned short)     0)
#define OBJECT    ((unsigned short)    01)
#define MONSTER   ((unsigned short)    02)
#define STAIRS    ((unsigned short)    04)
#define HORWALL   ((unsigned short)   010)
#define VERTWALL  ((unsigned short)   020)
#define DOOR      ((unsigned short)   040)
#define FLOOR     ((unsigned short)  0100)
#define TUNNEL    ((unsigned short)  0200)
#define TRAP      ((unsigned short)  0400)
#define HIDDEN    ((unsigned short) 01000)

#define ARMOR        ((unsigned short)   01)
#define WEAPON       ((unsigned short)   02)
#define SCROL        ((unsigned short)   04)
#define POTION       ((unsigned short)  010)
#define GOLD         ((unsigned short)  020)
#define FOOD         ((unsigned short)  040)
#define WAND         ((unsigned short) 0100)
#define RING         ((unsigned short) 0200)
#define AMULET       ((unsigned short) 0400)
#define ALL_OBJECTS  ((unsigned short) 0777)

#define LEATHER   0
#define RINGMAIL  1
#define SCALE     2
#define CHAIN     3
#define BANDED    4
#define SPLINT    5
#define PLATE     6
#define ARMORS    7

#define BOW               0
#define DART              1
#define ARROW             2
#define DAGGER            3
#define SHURIKEN          4
#define MACE              5
#define LONG_SWORD        6
#define TWO_HANDED_SWORD  7
#define WEAPONS           8

#define MAX_PACK_COUNT  24

#define PROTECT_ARMOR       0
#define HOLD_MONSTER        1
#define ENCH_WEAPON         2
#define ENCH_ARMOR          3
#define IDENTIFY            4
#define TELEPORT            5
#define SLEEP               6
#define SCARE_MONSTER       7
#define REMOVE_CURSE        8
#define CREATE_MONSTER      9
#define AGGRAVATE_MONSTER  10
#define MAGIC_MAPPING      11
#define CON_MON            12
#define SCROLS             13

#define INCREASE_STRENGTH   0
#define RESTORE_STRENGTH    1
#define HEALING             2
#define EXTRA_HEALING       3
#define POISON              4
#define RAISE_LEVEL         5
#define BLINDNESS           6
#define HALLUCINATION       7
#define DETECT_MONSTER      8
#define DETECT_OBJECTS      9
#define CONFUSION          10
#define LEVITATION         11
#define HASTE_SELF         12
#define SEE_INVISIBLE      13
#define POTIONS            14

#define TELE_AWAY       0
#define SLOW_MONSTER    1
#define INVISIBILITY    2
#define POLYMORPH       3
#define HASTE_MONSTER   4
#define MAGIC_MISSILE   5
#define CANCELLATION    6
#define DO_NOTHING      7
#define DRAIN_LIFE      8
#define COLD            9
#define FIRE           10
#define WANDS          11

#define STEALTH            0
#define R_TELEPORT         1
#define REGENERATION       2
#define SLOW_DIGEST        3
#define ADD_STRENGTH       4
#define SUSTAIN_STRENGTH   5
#define DEXTERITY          6
#define ADORNMENT          7
#define R_SEE_INVISIBLE    8
#define MAINTAIN_ARMOR     9
#define SEARCHING         10
#define RINGS             11

#define RATION  0
#define FRUIT   1

#define NOT_USED        ((unsigned short)   0)
#define BEING_WIELDED   ((unsigned short)  01)
#define BEING_WORN      ((unsigned short)  02)
#define ON_LEFT_HAND    ((unsigned short)  04)
#define ON_RIGHT_HAND   ((unsigned short) 010)
#define ON_EITHER_HAND  ((unsigned short) 014)
#define BEING_USED      ((unsigned short) 017)

#define NO_TRAP            -1
#define TRAP_DOOR           0
#define BEAR_TRAP           1
#define TELE_TRAP           2
#define DART_TRAP           3
#define SLEEPING_GAS_TRAP   4
#define RUST_TRAP           5
#define TRAPS               6

#define STEALTH_FACTOR  3
#define R_TELE_PERCENT  8

#define UNIDENTIFIED  ((unsigned short) 00) //MUST BE ZERO!
#define IDENTIFIED    ((unsigned short) 01)
#define CALLED        ((unsigned short) 02)

#define DROWS             24
#define DCOLS             80

#define MAX_TITLE_LENGTH  30
#define MAXSYLLABLES      40
#define MAX_METAL         14
#define WAND_MATERIALS    30
#define GEMS              14

#define GOLD_PERCENT  46

#define MAX_OPT_LEN  40

#define MAX_ID_TITLE_LEN  64

//The following #defines provide more meaningful names for some of the
//struct object fields that are used for monsters.  This, since each monster
//and object (scrolls, potions, etc) are represented by a struct object.
//Ideally, this should be handled by some kind of union structure.

#define m_damage           damage
#define hp_to_kill         quantity
#define m_char             ichar
#define first_level        is_protected
#define last_level         is_cursed
#define m_hit_chance       class
#define stationary_damage  identified
#define drop_percent       which_kind
#define trail_char         d_enchant
#define slowed_toggle      quiver
#define moves_confused     hit_enchant
#define nap_length         picked_up
#define disguise           what_is
#define next_monster       next_object

#define INIT_AW        (object*)0
#define INIT_RINGS     (object*)0
#define INIT_HP        12
#define INIT_STR       16
#define INIT_EXPLEVEL  1
#define INIT_EXP       0
#define INIT_PACK      {0}
#define INIT_GOLD      0
#define INIT_CHAR      '@'
#define INIT_MOVES     1250

#define MAXROOMS  9
#define BIG_ROOM  10

#define NO_ROOM  -1

#define PASSAGE  -3 //cur_room value

#define AMULET_LEVEL  26

#define R_NOTHING  ((unsigned short) 01)
#define R_ROOM     ((unsigned short) 02)
#define R_MAZE     ((unsigned short) 04)
#define R_DEADEND  ((unsigned short) 010)
#define R_CROSS    ((unsigned short) 020)

#define MAX_EXP_LEVEL  21
#define MAX_EXP        10000001L
#define MAX_GOLD       999999
#define MAX_ARMOR      99
#define MAX_HP         999
#define MAX_STRENGTH   99
#define LAST_DUNGEON   99

#define PARTY_TIME  10 //one party somewhere in each 10 level span

#define MAX_TRAPS  10  //maximum traps per level

#define HIDE_PERCENT 12

#define MONSTERS  26

#define HASTED          01L
#define SLOWED          02L
#define INVISIBLE       04L
#define ASLEEP          010L
#define WAKENS          020L
#define WANDERS         040L
#define FLIES           0100L
#define FLITS           0200L
#define CAN_FLIT        0400L //can, but usually doesn't, flit
#define CONFUSED        01000L
#define RUSTS           02000L
#define HOLDS           04000L
#define FREEZES         010000L
#define STEALS_GOLD     020000L
#define STEALS_ITEM     040000L
#define STINGS          0100000L
#define DRAINS_LIFE     0200000L
#define DROPS_LEVEL     0400000L
#define SEEKS_GOLD      01000000L
#define FREEZING_ROGUE  02000000L
#define RUST_VANISHED   04000000L
#define CONFUSES        010000000L
#define IMITATES        020000000L
#define FLAMES          040000000L
#define STATIONARY      0100000000L //damage will be 1,2,3,...
#define NAPPING         0200000000L //can't wake up for a while
#define ALREADY_MOVED   0400000000L

#define SPECIAL_HIT  (RUSTS|HOLDS|FREEZES|STEALS_GOLD|STEALS_ITEM|STINGS|DRAINS_LIFE|DROPS_LEVEL)

#define WAKE_PERCENT        45
#define FLIT_PERCENT        40
#define PARTY_WAKE_PERCENT  75

#define HYPOTHERMIA  1
#define STARVATION   2
#define POISON_DART  3
#define QUIT         4
#define WIN          5
#define KFIRE        6

#define UPWARD     0
#define UPRIGHT    1
#define RIGHT      2
#define DOWNRIGHT  3
#define DOWN       4
#define DOWNLEFT   5
#define LEFT       6
#define UPLEFT     7
#define DIRS       8

#define ROW1  7
#define ROW2  15

#define COL1  26
#define COL2  52

#define MOVED                 0
#define MOVE_FAILED           -1
#define STOPPED_ON_SOMETHING  -2
#define CANCEL                '\033'

#define HUNGRY  300
#define WEAK    150
#define FAINT   20
#define STARVE  0

#define MIN_ROW  1

#define HUNGER_STR_LEN 8



struct id
{
  short value;
  char title[MAX_ID_TITLE_LEN];
  char *real;
  unsigned short id_status;
};



struct obj
{                              //comment is monster meaning
  unsigned long m_flags;       //monster flags
  const char *damage;          //damage it does
  short quantity;              //hit points to kill
  short ichar;                 //'A' is for aquatar
  short kill_exp;              //exp for killing it
  short is_protected;          //level starts
  short is_cursed;             //level ends
  short class;                 //chance of hitting you
  short identified;            //'F' damage, 1,2,3...
  unsigned short which_kind;   //item carry/drop %
  short o_row, o_col, o;       //o is how many times stuck at o_row, o_col
  short row, col;              //current row, col
  short d_enchant;             //room char when detect_monster
  short quiver;                //monster slowed toggle
  short trow, tcol;            //target row, col
  short hit_enchant;           //how many moves is confused
  unsigned short what_is;      //imitator's charactor (?!%:
  short picked_up;             //sleep from wand of sleep
  unsigned short in_use_flags;
  struct obj *next_object;     //next monster
};

typedef struct obj object;



struct fightr
{
  object *armor;
  object *weapon;
  object *left_ring, *right_ring;
  short hp_current;
  short hp_max;
  short str_current;
  short str_max;
  object pack;
  long gold;
  short exp;
  long exp_points;
  short row, col;
  short fchar;
  short moves_left;
};

typedef struct fightr fighter;



struct dr
{
  short oth_room;
  short oth_row, oth_col;
  short door_row, door_col;
};

typedef struct dr door;



struct rm
{
  short bottom_row, right_col, left_col, top_row;
  door doors[4];
  unsigned short is_room;
};

typedef struct rm room;



struct tr
{
  short trap_type;
  short trap_row, trap_col;
};

typedef struct tr trap;



//////////////////////////////////////////////////////////////////////////////
//game.c
void game(void);
void player_init(void);

//hit.c
void mon_hit(object *monster);
void rogue_hit(object *monster, boolean force_hit);
void rogue_damage(short d, object *monster, short other);
int get_damage(const char *ds, boolean r);
int get_w_damage(const object *obj);
int get_number(const char *s);
long lget_number(const char *s);
int to_hit(const object *obj);
int damage_for_strength(void);
int mon_damage(object *monster, short damage);
void fight(boolean to_the_death);
void get_dir_rc(short dir, short *row, short *col, short allow_off_screen);
int get_hit_chance(const object *weapon);
int get_weapon_damage(const object *weapon);
void s_con_mon(object *monster);

//inventory.c
void inventory(const object *pack, unsigned short mask);
void mix_colors(void);
void make_scroll_titles(void);
void get_desc(const object *obj, char *desc);
void get_wand_and_ring_materials(void);
struct id *get_id_table(const object *obj);
void single_inv(short ch);
void id_type(void);

//level.c
void make_level(void);
void make_room(short rn, short r1, short r2, short r3);
int connect_rooms(short room1, short room2);
void clear_level(void);
void put_door(room *rm, short dir, short *row, short *col);
void draw_simple_passage(short row1, short col1, short row2, short col2, short dir);
int same_row(int room1, int room2);
int same_col(int room1, int room2);
void add_mazes(void);
void fill_out_level(void);
void fill_it(int rn, boolean do_rec_de);
void recursive_deadend(short rn, const short *offsets, short srow, short scol);
boolean mask_room(short rn, short *row, short *col, unsigned short mask);
void make_maze(short r, short c, short tr, short br, short lc, short rc);
void hide_boxed_passage(short row1, short col1, short row2, short col2, short n);
void put_player(short nr);
int drop_check(void);
int check_up(void);
void add_exp(int e, boolean promotion);
int get_exp_level(long e);
int hp_raise(void);
void show_average_hp(void);
void mix_random_rooms(void);

//monster.c
void put_mons(void);
object *gr_monster(object *monster, int mn);
void mv_mons(void);
void party_monsters(int rn, int n);
char gmc_row_col(int row, int col);
char gmc(object *monster);
void mv_1_monster(object *monster, short row, short col);
int mtry(object *monster, short row, short col);
void move_mon_to(object *monster, short row, short col);
int mon_can_go(const object *monster, short row, short col);
void wake_up(object *monster);
void wake_room(short rn, boolean entering, short row, short col);
const char *mon_name(const object *monster);
int rogue_is_around(int row, int col);
void wanderer(void);
void show_monsters(void);
void create_monster(void);
void put_m_at(short row, short col, object *monster);
void aim_monster(object *monster);
int rogue_can_see(int row, int col);
int move_confused(object *monster);
int flit(object *monster);
char gr_obj_char(void);
int no_room_for_monster(int rn);
void aggravate(void);
boolean mon_sees(const object *monster, int row, int col);
void mv_aquatars(void);

//move.c
int one_move_rogue(short dirch, short pickup);
void multiple_move_rogue(short dirch);
boolean is_passable(int row, int col);
boolean next_to_something(int drow, int dcol);
boolean can_move(int row1, int col1, int row2, int col2);
void move_onto(void);
boolean is_direction(short c, short *d);
boolean check_hunger(boolean msg_only);
boolean reg_move(void);
void rest(int count);
char gr_dir(void);
void heal(void);
boolean can_turn(short nrow, short ncol);
void turn_passage(short dir, boolean fast);

//object.c
void put_objects(void);
void put_gold(void);
void plant_gold(short row, short col, boolean is_maze);
void place_at(object *obj, int row, int col);
object *object_at(object *pack, short row, short col);
object *get_letter_object(int ch);
void free_stuff(object *objlist);
const char *name_of(const object *obj);
object *gr_object(void);
unsigned short gr_what_is(void);
void gr_scroll(object *obj);
void gr_potion(object *obj);
void gr_weapon(object *obj, int assign_wk);
void gr_armor(object *obj);
void gr_wand(object *obj);
void get_food(object *obj, boolean force_ration);
void put_stairs(void);
int get_armor_class(const object *obj);
void make_party(void);
void show_objects(void);
void put_amulet(void);
void rand_place(object *obj);
int r_index(const char *str, int ch, boolean last);
void c_object_for_wizard(void);

//pack.c
object *add_to_pack(object *obj, object *pack, int condense);
void take_from_pack(object *obj, object *pack);
object *pick_up(int row, int col, short *status);
void drop(void);
object *check_duplicate(object *obj, object *pack);
short next_avail_ichar(void);
short pack_letter(const char *prompt, unsigned short mask);
void take_off(void);
void wear(void);
void unwear(object *obj);
void do_wear(object *obj);
void wield(void);
void do_wield(object *obj);
void unwield(object *obj);
void call_it(void);
short pack_count(const object *new_obj);
boolean mask_pack(const object *pack, unsigned short mask);
boolean has_amulet(void);
void kick_into_pack(void);

//play.c
boolean is_digit(short ch);
void play_level(void);

//ring.c
void put_on_ring(void);
void do_put_on(object *ring, boolean on_left);
void remove_ring(void);
void un_put_on(object *ring);
void gr_ring(object *ring, boolean assign_wk);
void inv_rings(void);
void ring_stats(boolean pr);

//room.c
void light_up_room(int rn);
void light_passage(int row, int col);
void darken_room(short rn);
char get_dungeon_char(short row, short col);
char get_mask_char(unsigned short mask);
void gr_row_col(short *row, short *col, unsigned short mask);
short gr_room(void);
short party_objects(int rn);
short get_room_number(int row, int col);
boolean is_all_connected(void);
void visit_rooms(int rn);
void draw_magic_map(void);
void dr_course(object *monster, boolean entering, short row, short col);
boolean get_oth_room(short rn, short *row, short *col);

//save.c
long pseudorandom_seq(boolean init);
void save_game(void);
int restore_game(void);
void write_pack(const object *pack, FILE *fp);
void read_pack(object *pack, FILE *fp, boolean is_rogue);
void rw_dungeon(FILE *fp, boolean rw);
void rw_id(struct id id_table[], FILE *fp, int n, boolean wr);
void write_string(char *s, FILE *fp);
void read_string(char *s, FILE *fp, size_t len);
void rw_rooms(FILE *fp, boolean rw);
void r_read(FILE *fp, char *buf, int n);
void r_write(FILE *fp, const char *buf, int n);

//spec_hit.c
void special_hit(object *monster);
void rust(object *monster);
void freeze(object *monster);
void steal_gold(object *monster);
void steal_item(object *monster);
void disappear(object *monster);
void cough_up(object *monster);
boolean try_to_cough(short row, short col, object *obj);
boolean seek_gold(object *monster);
boolean gold_at(short row, short col);
void check_gold_seeker(object *monster);
boolean check_imitator(object *monster);
boolean imitating(short row, short col);
void sting(object *monster);
void drop_level(void);
void drain_life(void);
boolean m_confuse(object *monster);
boolean flame_broil(object *monster);
int get_dir(short srow, short scol, short drow, short dcol);

//throw.c
void throw(void);
boolean throw_at_monster(object *monster, object *weapon);
object *get_thrown_at_monster(object *obj, short dir, short *row, short *col);
void flop_weapon(object *weapon, short row, short col);
void rand_around(short i, short *r, short *c);

//trap.c
short trap_at(int row, int col);
void trap_player(short row, short col);
void add_traps(void);
void id_trap(void);
void show_traps(void);
void search(short n, boolean is_auto);

//use.c
void quaff(void);
void read_scroll(void);
void vanish(object *obj, short rm, object *pack);
void potion_heal(int extra);
void idntfy(void);
void eat(void);
void hold_monster(void);
void tele(void);
void hallucinate(void);
void unhallucinate(void);
void unblind(void);
void relight(void);
void take_a_nap(void);
void go_blind(void);
const char *get_ench_color(void);
void cnfs(void);
void unconfuse(void);
void uncurse_all(void);

//zap.c
void zapp(void);
object *get_zapped_monster(short dir, short *row, short *col);
void zap_monster(object *monster, unsigned short kind);
void tele_away(object *monster);
void wizardize(void);
void wdrain_life(object *monster);
void bounce(short ball, short dir, short row, short col, short r);

object *alloc_object(void);
void free_object(object *obj);

void win(void);
void quit(void);
boolean is_vowel(short ch);
void sell_pack(void);
int get_value(const object *obj);
void id_all(void);

void clear(void);
void mvaddch(int row, int col, int ch);
int mvinch(int row, int col);
//////////////////////////////////////////////////////////////////////////////
